import logging
import json
import requests

log = logging.getLogger(__name__)


def check_error(f):
    """Decorator provides consistent formatting for client http errors."""

    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except requests.exceptions.HTTPError as http_error:
            try:
                log.exception('Encountered an HTTP exception')
                data = http_error.response.json()
            except ValueError:
                data = http_error.response.text
            _response = {'error': True,
                         'code': http_error.response.status_code,
                         'data': data}
            log.debug('Response: %s', _response)
            return _response

    return wrapper


class InterfaceBase(object):
    """ Base HTTP Interface class """
    SERVICE_URI = ''

    def __init__(self, target, max_items=100, auth_token=None,
                 additional_headers=None, verify_ssl=True):
        """ Base Constructor

        :param target: URL and base URI of the target service
        :param max_items: The default maximum items to request for endpoints
        :param additional_headers: Additional headers to pass on, such as
        auth tokens.
        .. note:

            These can also be added for each request
        that support pagination.
        """
        self.target = target
        if self.SERVICE_URI:
            self.base_url = '{0}/{1}'.format(self.target, self.SERVICE_URI)
        else:
            self.base_url = self.target

        self.headers = {'Content-type': 'application/json'}

        if auth_token:
            self.headers.update({'X-Auth-Token': auth_token})
        if additional_headers:
            self.headers.update(additional_headers)

        self.max_items = max_items
        self.verify_ssl = verify_ssl

    def join_endpoint(self, endpoint):
        """
        :param endpoint:
        :return:
        """
        endpoint = endpoint or ''
        endpoint = endpoint.strip('/')
        if not endpoint:
            # endpoint is base_url
            return self.base_url
        return self.base_url + '/' + endpoint.lstrip('/')

    def get_per_request_headers(self, extra_request_headers):
        """

        :param extra_request_headers:
        :return:
        """
        if not extra_request_headers:
            return self.headers

        per_request_headers = self.headers.copy()
        per_request_headers.update(extra_request_headers)
        return per_request_headers

    @check_error
    def _request(self,
                 method,
                 item='',
                 data=None,
                 params=None,
                 extra_headers=None):
        """

        :param method:
        :param item:
        :param data:
        :param params:
        :param extra_headers:
        :return:
        """

        if data is not None:
            data = json.dumps(data)

        headers = self.get_per_request_headers(extra_headers)
        log.debug('Request: method=%s url=%s endpoint=%s params=%s data=%s '
                  'headers=%s', method, self.base_url, item, params, data,
                  headers)
        r = requests.request(method,
                             self.join_endpoint(item),
                             params=params,
                             data=data,
                             headers=headers,
                             verify=self.verify_ssl)
        r.raise_for_status()
        _response = r.json()
        log.debug('Response: %s', _response)
        return _response

    def get(self, item='', params=None, extra_headers=None):
        """

        :param item:
        :param params:
        :param extra_headers:
        :return:
        """
        return self._request('get', item, params=params,
                             extra_headers=extra_headers)

    def post(self, item='', data=None, params=None, extra_headers=None):
        """

        :param item:
        :param data:
        :param params:
        :param extra_headers:
        :return:
        """
        return self._request('post', item, data=data, params=params,
                             extra_headers=extra_headers)



import logging
import time

log = logging.getLogger(__name__)


class RPCException(Exception):
    pass


class Job(object):
    def __init__(self, rpc_client, query, instruction):
        """

        :param rpc_client:
        :param query:
        :param instruction:

        """
        self.rpc_client = rpc_client
        self.query = query
        self.instruction = instruction

        self.job_id = None

    @property
    def started(self):
        """ If a job_id is present, the job has been started"""
        return bool(self.job_id)

    def start(self):
        """ Start the job """
        r = self.rpc_client.post(data={
            'query': self.query,
            'instruction': self.instruction
        })

        if r.get('error'):
            if isinstance(r['data'], str):
                raise RPCException(r['data'])
            raise RPCException(r['data']['message'])

        self.job_id = r['job_id']

    @property
    def raw(self):
        """ Fetch the job from the server """
        if self.started:
            return self.rpc_client.get(self.job_id)

    @property
    def status(self):
        """ Get the jobs status endpoint """
        if self.started:
            return self.rpc_client.status(self.job_id)

    @property
    def tasks(self):
        """ Get the jobs tasks endpoint """
        if self.started:
            return self.rpc_client.tasks(self.job_id)

    @property
    def is_running(self):
        """ Check if the job is running or not """
        return self.started and not self.status['time_completed']

    def join(self, timeout=None, poll_interval=2):
        """

        :param timeout:
        :param poll_interval:
        :return: Tasks structure
        """
        if self.started:
            started = time.time()
            while True:
                if not self.is_running:
                    return self.status
                if timeout and time.time() - started > timeout:
                    break
                time.sleep(poll_interval)


class SimpleJob(Job):
    """ Creates an instruction in the standard form """
    def __init__(self, rpc_client, query, method, job_args=None,
                 job_kwargs=None):
        """

        :param rpc_client:
        :param query:
        :param method:
        :param job_args:
        :param job_kwargs:
        """
        self.method = method
        self.job_args = job_args or ()
        self.job_kw = job_kwargs or {}

        super(SimpleJob, self).__init__(rpc_client, query, instruction={
            'method': self.method,
            'args': self.job_args,
            'kwargs': self.job_kw
        })


class Preprocessor(Job):
    def __init__(self, rpc_client, query, preprocessor, payload):
        """

        :param rpc_client:
        :param query:
        :param preprocessor:
        :param payload:
        """
        self.preprocessor = preprocessor

        instruction = {'preprocessor': self.preprocessor}
        instruction.update(payload)
        super(Preprocessor, self).__init__(rpc_client, query, instruction)



from mercury_sdk.http.base import InterfaceBase


class QueryInterfaceBase(InterfaceBase):
    """ Used for endpoints that support /query"""

    @staticmethod
    def set_projection(params, projection):
        """

        :param params:
        :param projection:
        :return:
        """
        params.update({'projection': ','.join(projection)})

    def get(self, mercury_id=None, projection=None, params=None,
            extra_headers=None):
        """
        Override for get that add projection argument
        :param mercury_id:
        :param projection:
        :param params:
        :param extra_headers:
        :return:
        """

        params = params or {}

        if projection:
            self.set_projection(params, projection)
        return super(QueryInterfaceBase, self).get(item=mercury_id,
                                                   params=params,
                                                   extra_headers=extra_headers)

    @staticmethod
    def strip_empty(items):
        for item in items:
            for key in item:
                if isinstance(item[key], list):
                    item[key] = [el for el in item[key] if el]

    def query(self, query, item='/query', projection=None, params=None,
              extra_headers=None, strip_empty_elements=False):
        """

        :param query:
        :param item:
        :param projection:
        :param params:
        :param extra_headers:
        :param strip_empty_elements:
        :return:
        """

        params = params or {}

        if projection:
            self.set_projection(params, projection)

        data = self.post(item, data={'query': query}, params=params,
                         extra_headers=extra_headers)

        if strip_empty_elements:
            self.strip_empty(data['items'])

        return data



class Task(object):
    keys = [
        '_id',
        'action',
        'args',
        'backend',
        'host',
        'job_id',
        'kwargs',
        'mercury_id',
        'message',
        'method',
        'port',
        'progress',
        'status',
        'task_id',
        'time_completed',
        'time_started',
        'time_updated',
        'timeout',
        'traceback',
        'ttl_time_completed'
    ]

    def _create_attributes(self, kwargs):
        """
        Doing it this way means I can just pass **task into Task. Live long
        and prosper. (Downside is you lose tab complete in ipython/pycharm)
        :param kwargs:
        :return:
        """
        missing = []
        for key in self.keys:
            if key not in kwargs:
                missing.append(key)

        if missing:
            # This shouldn't ever happen
            raise ValueError('{} are missing. Shame.'.format(
                ','.join(missing)
            ))

        for key in self.keys:
            self.__setattr__(key, kwargs[key])

    def __init__(self, **kwargs):
        """

        :param kwargs:
        """
        self._create_attributes(kwargs)



from mercury_sdk.http.base import InterfaceBase


class JobInterfaceBase(InterfaceBase):
    SERVICE_URI = 'api/rpc/jobs'

    def get(self, job_id='', params=None, extra_headers=None):
        """

        :param job_id:
        :param params:
        :param extra_headers:
        :return:
        """
        return super(JobInterfaceBase, self).get(item=job_id, params=params,
                                                 extra_headers=extra_headers)

    def status(self, job_id):
        """

        :param job_id:
        :return:
        """
        return self.get('/{}/status'.format(job_id))

    def tasks(self, job_id):
        """

        :param job_id:
        :return:
        """
        return self.get('/{}/tasks'.format(job_id))

    def submit(self, query, instruction):
        """

        :param query:
        :param instruction:
        :return:
        """
        return self.post(item=None,
                         data={'query': query, 'instruction': instruction})
